from __future__ import (
    absolute_import,
    unicode_literals,
)

from typing import (  # noqa: F401 TODO Python 3
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    TypeVar,
    Union,
)

import attr
import six


__all__ = (
    'ActionRequest',
    'ActionResponse',
    'Body',
    'Context',
    'Control',
    'JobRequest',
    'JobResponse',
    'Error',
    'UnicodeKeysDict',
)


_VT = TypeVar('_VT')
_AT = TypeVar('_AT')


class UnicodeKeysDict(dict):
    """
    A special dictionary factory used to ensure that Attrs object-to-dict dictionaries contain keys that are only
    unicode strings instead of byte strings. This factory will become unnecessary and be removed in PySOA 2.0.0 when
    all Python 2 support is removed.
    """

    def __setitem__(self, key, value):  # type: (six.text_type, Any) -> None
        super(UnicodeKeysDict, self).__setitem__(six.text_type(key), value)

    def setdefault(self, key, default=None):  # type: (six.text_type, Optional[_VT]) -> _VT
        return super(UnicodeKeysDict, self).setdefault(six.text_type(key), default)


Body = Dict[six.text_type, Any]
"""A type used for annotating attributes and arguments that represent action request and response bodies."""

Context = Dict[six.text_type, Any]
"""A type used for annotating attributes and arguments that represent job request and response context headers."""

Control = Dict[six.text_type, Any]
"""A type used for annotating attributes and arguments that represent job request control headers."""


@attr.s(frozen=True)
class Error(object):
    """
    Represents an error that occurred, in the format transmitted over the transport between client and service.

    :param code: The machine-readable error code.
    :param message: The human-readable error message.
    :param field: If the error is the result of validation of a job attribute or action request field, this contains
                  the name of that attribute or field. It will be `None` otherwise.
    :param traceback: If the error is the result of an exception with valuable traceback information, this contains
                      that traceback. It will be `None` otherwise.
    :param variables: If there are any variables pertinent to this error, this dictionary will contain the names and
                      values of those variables. It will be `None` otherwise. New in version 0.9.0.
    :param denied_permissions: If this error is the result of insufficient privileges to perform the requested
                               operation, this attribute will be a list containing the necessary permissions that
                               were denied, if the service supports reporting this information.  It will be `None`
                               otherwise. New in version 0.44.0.
    :param is_caller_error: Indicates whether this error is the result of the caller doing something wrong (`True`),
                            such as an invalid job request or an action request body that fails to validate, or the
                            result of a server or service problem or bug (`False`). New in version 0.70.0.
    """

    code = attr.ib()  # type: six.text_type
    message = attr.ib()  # type: six.text_type
    field = attr.ib(default=None)  # type: Optional[six.text_type]
    traceback = attr.ib(default=None)  # type: Optional[six.text_type]
    variables = attr.ib(default=None)  # type: Dict[six.text_type, Any]
    denied_permissions = attr.ib(default=None)  # type: Optional[List[six.text_type]]
    is_caller_error = attr.ib(default=False, metadata={'added_in_version': (0, 70, 0)})  # type: bool


def _convert_errors(errors):
    # type: (Union[Iterable[Mapping[six.text_type, Any]], Iterable[Error]]) -> List[Error]
    value = []  # type: List[Error]
    for a in errors:
        value.append(a if isinstance(a, Error) else Error(**a))
    return value


@attr.s
class ActionRequest(object):
    """
    A request that the server execute a single action.

    :param action: The name of the action to execute.
    :param body: The request body input.
    """

    action = attr.ib()  # type: six.text_type
    body = attr.ib(default=attr.Factory(dict))  # type: Body


def _convert_action_requests(actions):
    # type: (Union[Iterable[Mapping[six.text_type, Any]], Iterable[ActionRequest]]) -> List[ActionRequest]
    value = []  # type: List[ActionRequest]
    for a in actions:
        value.append(a if isinstance(a, ActionRequest) else ActionRequest(**a))
    return value


@attr.s
class ActionResponse(object):
    """
    A response generated by a single action on the server.

    :param action: The name of the action that was executed.
    :param body: The response body output.
    :param errors: A list of any action :class:`Error` or errors that occurred, or an empty list if no errors occurred.
    """

    action = attr.ib()  # type: six.text_type
    errors = attr.ib(default=attr.Factory(list), converter=_convert_errors)  # type: List[Error]
    body = attr.ib(default=attr.Factory(dict))  # type: Body


def _convert_action_responses(actions):
    # type: (Union[Iterable[Mapping[six.text_type, Any]], Iterable[ActionResponse]]) -> List[ActionResponse]
    value = []  # type: List[ActionResponse]
    for a in actions:
        value.append(a if isinstance(a, ActionResponse) else ActionResponse(**a))
    return value


@attr.s
class JobRequest(object):
    """
    A request that the server execute a job.

    A job consists of one or more actions and context and control headers. Each action is an :class:`ActionRequest`,
    while the context and control headers are dictionaries.

    :param control: The control header dictionary.
    :param context: The context header dictionary.
    :param actions: The list of `ActionRequest` objects to be executed. This list must contain at least one request.
    """

    control = attr.ib(default=attr.Factory(dict))  # type: Control
    context = attr.ib(default=attr.Factory(dict))  # type: Context
    actions = attr.ib(default=attr.Factory(list), converter=_convert_action_requests)  # type: List[ActionRequest]


@attr.s
class JobResponse(object):
    """
    A response generated by a server job.

    Contains the result or error generated by each action in the job.

    :param context: The context header dictionary.
    :param actions: The list of :class:`ActionResponse` objects that were executed. This list *may* be shorter than the
                    number of actions in the :class:`JobRequest` (even empty) if the `errors` attribute contains
                    one or more errors.
    :param errors: A list of any job :class:`Error` or errors that occurred, or an empty list of no errors occurred.
    """

    errors = attr.ib(default=attr.Factory(list), converter=_convert_errors)  # type: List[Error]
    context = attr.ib(default=attr.Factory(dict))  # type: Context
    actions = attr.ib(default=attr.Factory(list), converter=_convert_action_responses)  # type: List[ActionResponse]
